# frozen_string_literal: true

# https://www.wrike.com/open.htm?id=461636692
class CloseContractJob < ActiveJob::Base

  queue_as :closecontract
  unique :until_executed
  
  def perform(entity_validation_id)
    # Get entity validation
    entity_validation               = Hermes::EntityValidation.find(entity_validation_id)
    # Get applicant requests
    applicant_housing_request_ids   = Hermes::HousingRequest.where(main_applicant_id: entity_validation.applicant.id).map(&:id)
    # Get entity offers
    entity_housing_offer_ids        = entity_validation.entity.housing_offer_ids
    # Get applicant and entity linked applications
    housing_applications = Hermes::HousingApplication.where(
      housing_offer_id: { :$in => entity_housing_offer_ids },
      housing_request_id: { :$in => applicant_housing_request_ids }
    )
    # Get requests from linked applications
    housing_requests = housing_applications.map(&:housing_request)
    # Remove entity_id from linked requets
    housing_requests.each do |entity_applicant_request|
      entity_applicant_request.main_applicant.update(main_employer_siret: nil)
    end
    # System cancel linked applications
    housing_applications.each(&:system_cancel_application!)
    # Return entity validation
    entity_validation.reload
  end
end
