# frozen_string_literal: true

# Legende
# Les script a plusieurs parties indique que les bash (Batch dans l'app)
# peuvent s'initier avec differents types de source
# Le besoin est donc d'isoler la logique associee au modele actuel
# ne pas oublier d'indexer les champs simples

module Hermes
  # Housing offer model
  class HousingOffer
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::AuditableConcern
    include Hermes::Concerns::AlsIdConcern
    include Hermes::Concerns::Status
    include Hermes::Concerns::ReservationHolder
    include ::Concerns::Geolocalizable
    include Hermes::Concerns::SopraXmlSender
    
    include Mongoid::Locker
    index({locking_name: 1}, name: 'mongoid_locker_index', partial_filter_expression: {locking_name: {:$exists => true}}, expire_after_seconds: lock_timeout)
    
    hermes_tools 'typology'
    
    RENTED_STATUSES = %w[LOUE loué loue PLACE placé place].freeze
    PUBLISHED_STATUSES = ['LOGEMENT PUBLIE', 'PUBLIE'].freeze
    ATTRIBUTION_STATUSES = %w[requalification REQUALIFICATION]
    
    POSSIBLE_ELIGIBILITY_STATUSES = [
      "PUBLIE",
      "LOGEMENT PUBLIE",
      "EN COURS DE DESIGNATION",
      "EN COURS D'ATTRIBUTION",
      "RESERVE ALS",
      "EN ATTENTE COMMERCIALISATION",
      "COMMERCIALISE SANS PUBLICATION",
    ].freeze
    
    ORIENTATION_BOOLS = %i[
      priority_public
      single_parent
      mutation
    ].freeze
    
    
    field :requalification, type: Boolean, default: false
    field :dalo, type: Boolean
    field :entry_date, type: DateTime
    field :external_ref, type: String
    field :communicate_exact_address, type: Boolean
    field :typology, type: String
    field :type, type: String
    field :age, type: String
    field :department, type: String
    field :publication_end_date, type: DateTime
    field :max_applicants, type: Integer
    field :minimal_publication_days, type: Integer, default: 3
    field :expiry_days, type: Integer, default: 7
    field :rent_amount, type: Float
    field :rent_with_charges, type: Float
    field :surface, type: Integer
    field :rooms, type: Integer
    field :exclusivity_end_date, type: DateTime
    field :nb_of_eligible_employees, type: Integer
    field :nb_of_requests_from_employees, type: Integer
    field :availability_date, type: DateTime
    field :description, type: String
    field :details, type: String
    field :effort_rate, type: Float
    field :min_effort_treshold, type: Float
    field :single_parent, type: Boolean
    field :reserved, type: Boolean
    field :reservation_agreement_ref, type: String
    field :priority_public, type: Boolean
    field :mobility, type: Boolean
    field :bedrooms, type: Integer
    field :bathrooms, type: Integer
    field :floor, type: Integer
    field :floors, type: Integer
    field :elevators, type: Integer
    field :has_elevator, type: Boolean
    field :has_gardian, type: Boolean
    field :outside, type: String
    field :cellars, type: Integer
    field :has_cellar, type: Boolean
    field :parking_type, type: String
    field :parking_charges, type: Float, default: 0.0
    field :pmr, type: Boolean
    field :kitchen_type, type: String
    field :construction_year, type: Date
    field :address, type: String
    field :address_complements, type: String
    field :mutation, type: Boolean
    field :eligibility_query, type: String
    field :threshold_type, type: String
    field :threshold_subtype, type: String
    field :heating_type, type: String
    field :hot_water_distribution, type: String
    field :individual_heating, type: Boolean
    field :date_publication_start, type: DateTime
    field :date_publication_end, type: DateTime
    field :kind, type: String
    field :building_type, type: String
    field :already_rent, type: Boolean
    field :guarantee_deposit, type: Float
    field :rental_charges, type: Float, default: 0.0
    field :reference, type: String
    field :release_date, type: Date
    field :dpe_ges, type: String
    field :dpe_conso, type: String
    
    field :ceil_category_1, type: Float
    field :ceil_category_2, type: Float
    field :ceil_category_3, type: Float
    field :ceil_category_4, type: Float
    field :ceil_category_5, type: Float
    field :ceil_category_6, type: Float
    field :ceil_category_7, type: Float
    field :ceil_category_8, type: Float
    field :ceil_category_9, type: Float
    field :ceil_category_10, type: Float
    field :ceil_category_11, type: Float
    field :ceil_category_12, type: Float
    field :floor_category_1, type: Float
    field :floor_category_2, type: Float
    field :floor_category_3, type: Float
    field :floor_category_4, type: Float
    field :floor_category_5, type: Float
    field :floor_category_6, type: Float
    field :floor_category_7, type: Float
    field :floor_category_8, type: Float
    field :floor_category_9, type: Float
    field :floor_category_10, type: Float
    field :floor_category_11, type: Float
    field :floor_category_12, type: Float
    field :min_revenues, type: Float
    field :total_charges, type: Float
    field :min_effort_rate, type: Float, default: 0.333
    field :tense_area, type: Boolean
    field :subsidiary_company, type: String
    field :rpls_id, type: String
    field :operation_name, type: String
    field :administrator, type: String
    field :administration_area, type: String
    field :potential_request_name_town, type: String
    field :potential_request_name_county, type: String
    field :number_of_companies_30_minutes_around, type: Integer
    field :construction_type, type: String
    field :address_can_be_shared, type: String
    field :qpv, type: Boolean
    field :total_floors, type: String
    field :address_extended, type: String
    field :test_created_at, type: DateTime
    field :already_inhabited, type: Boolean
    field :new_offer, type: Boolean
    field :lessor_name, type: String
    field :lessor_siren, type: String
    field :als_siren, type: String
    field :postal_code, type: String
    field :insee_code, type: String
    field :residence_code, type: String
    field :reserving_business_name, type: String
    field :reserving_siren, type: String
    field :residence_title, type: String
    # field :main_picture_id, type: BSON::ObjectId
    # field :pictures, type: String
    field :agency_id, type: BSON::ObjectId
    field :agency_name, type: String
    field :lessor_contact_name, type: String
    field :lessor_email, type: String
    field :lessor_phone_number, type: String
    field :terraces_or_balconies, type: Integer
    field :gardens, type: Integer
    field :has_garden, type: Boolean
    field :green_area, type: Boolean
    field :condo, type: Boolean
    field :furnished_lodging, type: Boolean
    field :virtual_visit, type: Boolean
    field :duplex, type: Boolean
    # TODO : verifier le match au noveau du WIKI
    field :first_quartile, type: String
    field :other_quartiles, type: String
    field :lessor_convention_reference, type: String
    field :reserving_agency_name, type: String
    field :reserving_agency_siren, type: String
    field :is_public, type: Boolean
    field :published, type: Boolean
    field :accept_low_occupation, type: Boolean, default: false
    field :designation_date_end, type: DateTime
    field :designation_period_duration, type: Integer
    field :lessor_orientation, type: String
    field :disability_accesses, type: Integer
    field :has_disability_access, type: Boolean
    field :eligible_requests_nb, type: Integer
    field :possible_eligible_requests_nb, type: Integer
    field :applicated_nb, type: Integer, default: 0
    field :application_following_contact, type: String
    field :entities_siret, type: Array
    field :eligible_requests_nb_detailed, type: Hash # of form { priority_public: 23, mutation: 45, ... }
    field :commercialized, type: Boolean, default: false
    field :entity_ids, type: Array, default: []
    field :active_reservations, type: Array, default: []
    field :active_reservations_rex_ids, type: Array, default: []
    # associated pictures
    field :ordered_picture_ids, type: Array
    field :picture_urls, type: Array, default: []
    field :is_mad, type: Boolean, default: false
    field :is_mad_updated_at, type: DateTime
    field :mad_als_date, type: DateTime
    
    field :max_family_size, type: Integer, default: 0
    field :couple_min_family_size, type: Integer, default: 0
    field :single_min_family_size, type: Integer, default: 0
    field :last_update_by_lessor, type: DateTime
    field :auto_destroy_next_day, type: Boolean
    field :is_from_form, type: Boolean
    field :housing_channel_keys, type: Array, default: []
    field :reference_mad, type: String, default: ""
    field :delai_reponse, type: Integer, default: 30
    
    # validates :entities_siret_length
    validates :lessor_siren, length: {is: 9}, allow_blank: true
    validates :als_siren, length: {is: 9}, allow_blank: true
    
    validates :reserving_siren, length: {is: 9}, allow_blank: true
    validates :reserving_agency_siren, length: {is: 9}, allow_blank: true
    validate :picture_urls_format
    
    alias_attribute :district, :city
    
    common_indexes *%i[
      entry_date
      external_ref
      typology
      publication_end_date
      rent_amount
      date_publication_start
      min_revenues
      postal_code
      agency_id
      lessor_siren
      published
      lessor_id
      entity_ids
      mutation
      mobility
      priority_public
      single_parent
      rent_with_charges
      is_public
      single_min_family_size
      couple_min_family_size
      max_family_size
      department
      threshold_type
      threshold_subtype
      active_reservations
      active_reservations_rex_ids
      reserved
      couple_min_family_size
      ceil_category_1
      ceil_category_2
      ceil_category_3
      ceil_category_4
      ceil_category_5
      ceil_category_6
      floor_category_1
      floor_category_2
      floor_category_3
      floor_category_4
      floor_category_5
      floor_category_6
      housing_channel_keys
    ]
    
    attr_accessor :will_update_lessor_objectives, :minimal_publication_date, :f5_sended
    
    belongs_to :threshold, class_name: 'Hermes::Threshold', index: true
    belongs_to :lessor, class_name: 'Hermes::Lessor', index: true
    # TODO: handle also with_entity_ids=
    has_many :housing_entity_offers, class_name: 'Hermes::HousingEntityOffer'
    has_and_belongs_to_many :target_objectives, class_name: 'Hermes::LessorObjective'
    has_many :scores, class_name: 'Hermes::Score'
    has_many :housing_entity_reservations, class_name: 'Hermes::EntityReservation'
    has_many :pictures, class_name: 'Hermes::Picture'

    
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    
    reader_auth_roles 'collaborateur offre logement'
    
    before_save :before_save_actions # https://dev.hermes.openstack.local/admin/families/5cffb7af4ac06e63bbd93298/mapping
    after_create :after_create_actions
    after_save :after_save_actions
    after_update :after_update_actions
    
    ELIGIBILITY_FIELDS = %w[
      entity_ids
      mutation
      mobility
      offer_status
      priority_public
      single_parent
      rent_with_charges
      is_public
      single_min_family_size
      couple_min_family_size
      max_family_size
      min_revenues
      department
      threshold_type
      threshold_subtype
      active_reservations
      housing_channel_keys
    ].concat(
      (1..12).each_with_object([]) do |i, arr|
        %w[ceil floor].each {|type| arr << "#{type}_category_#{i}"}
      end
    ).freeze
    
    THRESHOLDS_FIELDS = %w[
      threshold_type
      threshold_subtype
    ].freeze
    
    HousingOffer.add_status(:single, 'lessor_status')
    HousingOffer.add_status(:single, 'offer_status')
    HousingOffer.add_status(:multiple, 'entity_status')
    
    scope :will_trigger_objective, ->(objective_id) {
      where(target_objective_ids: objective_id, :lessor_status.nin => RENTED_STATUSES, :offer_status.nin => ATTRIBUTION_STATUSES)
    }
    
    ###############################
    #### Notification #############
    include ::Concerns::NotificationCenter::Notifiable
    
    def self.targets_selector(housing_offer, opt = {})
      housing_offer.reload
      applicant_ids = opt.fetch(:target_ids, nil)
      # applicant_ids = Hermes::Score.where(housing_offer_id: housing_offer.id).map(&:applicant_id)
      eligibility = housing_offer.eligibility_query ? housing_offer.formated_eligibility : {}
      applicant_ids ||= Hermes::HousingRequest.where(is_active: true, **eligibility).pluck(:main_applicant_id).map(&:to_s)
      Hermes::Applicant.where(:_id.in => applicant_ids || [])
    end
    
    def notification_mailer
      ::Hermes::NotificationCenter::HermesMailer
    end
    
    def notifiable_title(key, opt = {})
      I18n.t("#{key}.title", reference: self.id)
    end
    
    def notifiable_body(key, opt = {})
      I18n.t("#{key}.body")
    end
    
    def udapte_siret_siren
      self.lessor_siren = self.lessor_siren&.slice(0, 9)
      self.als_siren = self.als_siren&.slice(0, 9)
      self.reserving_siren = self.reserving_siren&.slice(0, 9)
      self.reserving_agency_siren = self.reserving_agency_siren&.slice(0, 9)
    end
    
    def notifiable_path
      "https://als-hermes-int.be-idys.com/housing_offers/#{id}"
    end
    
    ###############################
    ###############################
    
    def compute_effort_rate
      typo_sym = typology&.upcase
      if HERMES_CONFIG["EFFORT_RATES"] && HERMES_CONFIG["EFFORT_RATES"][typo_sym].present?
        self.effort_rate = HERMES_CONFIG["EFFORT_RATES"][typo_sym]
      else
        self.effort_rate = (self.min_effort_rate || 0)
      end
    end
    
    def construction_year=(string)
      super(Date.new(Integer(string))) if string
    end
    
    def contact_group
      lessor&.contact_group
    end
    
    def main_contact
      contact_group&.main_contact
    end
    
    def entities
      Hermes::Entity.in(id: housing_entity_offers.pluck(:entity_id))
    end
    
    def entity_ids
      entities.pluck(:id)
    end
    
    def entity_ids=(*arr)
      previous_ids = entities.pluck(:id).map(&:to_s)
      new_ids = (arr || []).map(&:to_s)
      (new_ids - previous_ids).each do |entity_id|
        Hermes::HousingEntityOffer.create(entity_id: entity_id, housing_offer_id: id)
      end
      Hermes::HousingEntityOffer.where(:entity_id.in => (previous_ids - new_ids), housing_offer: id).destroy
      super(entity_ids)
    end
    
    def lessor_id=(lessor_new_id)
      return unless lessor_id != id
      
      Hermes::HousingEntityOffer.find_by(entity_id: lessor_id, housing_offer_id: id)&.destroy
      Hermes::HousingEntityOffer.create(entity_id: lessor_new_id, housing_offer_id: id)
      super(lessor_new_id)
    end
    
    def agency
      Hermes::Entity.find_by(id: agency_id)
    end
    
    def history
      Hermes::LessorObjectiveHistory.where(housing_offer_id: self.id)
    end
    
    def detailed_score
      Hermes::Score.where(housing_offer_id: self.id)
    end
    
    def inherit_lesssor_data
      return unless lessor
      
      self.lessor_name = lessor.title
    end
    
    def update_extras_presence
      %w[elevator garden cellar disability_access].each do |extra_name|
        send("has_#{extra_name}=", send(extra_name.pluralize)&.send('>', 0))
      end
    end
    
    # set relationship with threshold
    def update_own_thresholds
      return unless threshold
      
      threshold.set_offer_thresholds(self)
    end
    
    def update_eligbility_old
      # https://dev.hermes.openstack.local/admin/batchs/5d2301804ac06e2ce9126715/script
      # script tri-partie (offer || (applicant || requester))
      if target = (filterer = FilterGenerator.new(context.select {|key| %w[offer_id applicant_id request_id].include? key})).target_instance
        new_sfs = target.superfields
        new_sfs[filterer.process_watch_field] = 'off'
        if target.set(superfields: new_sfs, updated_at: DateTime.current) && filterer.set_filter_on_target
          Rails.logger.debug "\n BASH process ended with filters #{filterer.filters} at #{target.reload.superfields[filterer.process_watch_field]}\n\n"
          "Elligibility on #{filterer.target} passed with result : #{filterer.filters}"
        else
          false
        end
      else
        Rails.logger.error "Element #{context.each_with_object([]) do |(_key, value), arr|
          arr.push value if value.present?
        end.first} could not be found"
        false
      end
    end
    
    # integration olivier => batch https://dev.hermes.openstack.local/admin/families/5cffb7af4ac06e63bbd93298/mapping-
    def update_family_size
      surface = self.surface || 0
      typologint = case typology
                   when /^T([1-9][0-9]*?)$/
                     Integer(::Regexp.last_match(1))
                   when 'Chambre'
                     1
                   when 'T6 et plus'
                     rooms
                   end
      typologint = (typologint&.positive? ? typologint : rooms) || 0
      min_family_size = (min_calculated = typologint - 2) < 1 ? 1 : min_calculated
      max_size = typologint < 6 ? (typologint + 1).to_s : 24
      single_min_dependents = 0
      i = 0
      surface_bis = surface || 0
      while surface_bis.positive?
        can_increment = surface_bis -= if i == 1
                                         7
                                       elsif i < 8
                                         9
                                       else
                                         8
                                       end
        single_min_dependents += if (i.zero? && can_increment > 6) || (i.between?(1, 7) && can_increment > 8) || (i > 7 && can_increment > 7)
                                   1
                                 elsif i.zero? && can_increment.negative?
                                   -1
                                 else
                                   0
                                 end
        i += 1
      end
      s_min_family_size = [min_family_size, single_min_dependents].min - ((accept_low_occupation == true ? 1 : 0) * 1)
      couple_min_family_size = ((surface - 16) / 9).floor + 2 - ((accept_low_occupation === true ? 1 : 0) * 1)
      c_min_family_size = [min_family_size, couple_min_family_size].min - ((accept_low_occupation == true ? 1 : 0) * 1)
      int_max_size = max_size.is_a?(String) ? Integer(max_size.sub('+', '')) : max_size
      self.single_min_family_size = s_min_family_size > int_max_size ? int_max_size : s_min_family_size
      self.couple_min_family_size = c_min_family_size > int_max_size ? int_max_size : c_min_family_size
      self.max_family_size = max_size
    end
    
    def update_individual_heating
      self.individual_heating = heating_type && heating_type =~ /^Individuel/
    end
    
    def update_charges
      tc = (rental_charges || 0.0) + (parking_charges || 0.0) + calculate_utility_charges
      self.total_charges = tc
    end
    
    def compute_without_charges
      # Nous supposons que si le loyer hors charges n'est pas indiqué, 
      # cela signifie que le montant des charges est nul 
      # et que le montant du loyer avec charges et le loyer hors charges sont égaux.
      self.rent_amount = rent_with_charges
    end
    
    def calculate_utility_charges
      return 0 if typology.blank? || heating_type.blank? || !['T1', 'T2', 'T3', 'T4', 'T5', 'T6'].include?(typology)
      
      non_electric = {
        'T1' => [31.59, 10.15],
        'T2' => [46.86, 15.06],
        'T3' => [62.84, 20.20],
        'T4' => [77.03, 24.76],
        'T5' => [104.58, 33.61],
        'T6' => [144.22, 45.71],
      }
      
      electric = {
        'T1' => [37.71, 11.38],
        'T2' => [55.09, 16.63],
        'T3' => [74.05, 22.29],
        'T4' => [90.40, 27.22],
        'T5' => [122.27, 36.81],
        'T6' => [166.29, 50.06],
      }
      
      if ['Individuel gaz', 'Individuel PAC Electrique', 'Collectif autre', 'Individuel autre'].include?(heating_type)
        return non_electric[typology].inject(:+)
      end
      
      if ['Individuel électrique'].include?(heating_type)
        return electric[typology].inject(:+)
      end
      0
    end
    
    def update_min_revenues
      formated_min_effort_rate = (min_effort_rate&.to_f || 0.0)
      self.min_revenues = (formated_min_effort_rate.positive? ? 1 / formated_min_effort_rate : 0.0) * (rent_with_charges || 0.0)
    end
    
    def update_threshold
      self.threshold = Hermes::Threshold.find_by("$or" => [{type: threshold_type}, {codes: threshold_type}])
    end
    
    def formated_eligibility
      return if eligibility_query.blank?
      JSON.parse(eligibility_query).each_with_object({}) do |(key, val), h|
        h[key.to_sym] = val unless val.nil?
      end
    end
    
    def update_eligible_requests_nb_detail
      self.eligible_requests_nb_detailed ||= {}
      ORIENTATION_BOOLS.each do |orientation_name|
        orientation = send(orientation_name)
        next unless orientation
        
        self.eligible_requests_nb_detailed[orientation_name] = eligible_requests.where(orientation_name.to_sym => orientation).count
      end
      
      if reserved && entity_ids
        self.eligible_requests_nb_detailed[entities] = entity_ids.each_with_object({}) do |entity_id, h|
          h[entity_id] = eligible_requests.where(entity_id: entity_id).count
        end
      end
    end

    def count_eligible_request_nb
      return 0 unless formated_eligibility
      query = formated_eligibility.merge(is_active: true, :request_status.in => ['ELIGIBLE', 'COMPLETE'])
      Hermes::HousingRequest.where(query).count
    end 

    def count_possible_eligible_request_nb
      return 0 unless formated_eligibility
      query = formated_eligibility.merge(is_active: true)
      Hermes::HousingRequest.where(query).count
    end 

    def zero_eligible
      is_public || formated_eligibility.nil?
    end 
    
    def update_eligible_request_nb
      self.eligible_requests_nb = zero_eligible ? 0 : count_eligible_request_nb
      self.possible_eligible_requests_nb = zero_eligible ? 0 : count_possible_eligible_request_nb
    end

    def calculate_eligible_requests_nb!
      self.set(
        updated_at: DateTime.current, 
        eligible_requests_nb: zero_eligible ? 0 : count_eligible_request_nb,
        possible_eligible_requests_nb: zero_eligible ? 0 : count_possible_eligible_request_nb,
      )
    end
    
    def update_department
      self.department = (
      begin
        postal_code[0..1].to_s
      rescue StandardError
        nil
      end)
    end
    
    def update_is_mad_updated_at
      self.designation_date_end = DateTime.current + 30.days
      self.is_mad_updated_at = DateTime.current
    end
    
    def handle_unpublished_status
      if PUBLISHED_STATUSES.include?(offer_status) && date_publication_start?
        expiration = self.expiry_days
        setting = Hermes::HousingOfferExpirationSetting.find_by(department: department, lessor_id: lessor_id)
        setting = Hermes::HousingOfferExpirationSetting.find_by(department: department, lessor_id: nil) if setting.nil?
        if setting
          expiration = setting.expiry_days
          set(
            max_applicants: setting.max_applicants,
            minimal_publication_days: setting.minimal_publication_days,
            updated_at: DateTime.current
          )
        end
        
        @minimal_publication_date = publication_end_date.presence || (date_publication_start + expiration)
        
        if !max_applicants.nil? && max_applicants <= applicated_nb
          @minimal_publication_date = date_publication_start + minimal_publication_days
          if Time.zone.now > @minimal_publication_date
            set(publication_end_date: Time.zone.now, offer_status: 'COMMERCIALISATION NON PUBLIEE', updated_at: DateTime.current)
            return
          end
        end
        
        set(publication_end_date: @minimal_publication_date, updated_at: DateTime.current)
        UnpublishHousingOfferJob.set(wait_until: @minimal_publication_date).perform_later(self.id.to_s)
      end
    end
    
    def reset_offers
      Hermes::HousingOffer.all.each_with_index do |off, i|
        lessor_siren1 = off.lessor_siren
        off.lessor_siren = nil
        off.save
        off.lessor_siren = lessor_siren1
        off.save
      end
    end
    
    def handle_entities_relationships
      formated_siret = entities_siret || []
      formated_siren = []
      formated_siret.each {|element| formated_siren.push element.slice(0, 9)}
      formated_siret_was = entities_siret_was || []
      formated_siren_was = []
      formated_siret_was.each {|element| formated_siren_was.push element.slice(0, 9)}
      formated_siren_plus_lessor = (formated_siren + [lessor_siren]).compact
      formated_siren_was_plus_lessor = (formated_siren_was + [lessor_siren_was]).compact
      
      all_removed_entities = Hermes::Entity.where(:siren.in => formated_siren_was_plus_lessor - formated_siren_plus_lessor)
      all_added_entities = Hermes::Entity.where(:siren.in => formated_siren_plus_lessor - formated_siren_was_plus_lessor)
      
      not_lessor = {siren: {:$ne => lessor_siren}}
      removed_entities = all_removed_entities.where(not_lessor)
      added_entities = all_added_entities.where(not_lessor)
      Hermes::HousingEntityOffer.where(
        :entity_id.in => removed_entities.pluck(:id),
        housing_offer_id: id
      ).set(
        is_active: false,
        updated_at: DateTime.current
      )
      
      added_entities.each do |entity|
        offer_entity = Hermes::HousingEntityOffer.find_or_create_by(housing_offer_id: id, entity_id: entity.id)
        offer_entity.is_active = true
        offer_entity.save
      end
      
      # TODO: better handling for groups
      all_removed_entities.each do |entity|
        readers.delete("g_#{entity.supervisor_group.id}")
        writers.delete("g_#{entity.manager_group.id}")
      end
      all_added_entities.each do |entity|
        readers.push("g_#{entity.supervisor_group.id}") if entity.supervisor_group.present?
        writers.push("g_#{entity.manager_group.id}")    if entity.manager_group.present?
      end
    end
    
    def handle_is_mad
      return if is_mad == true
      
      self.is_mad = (offer_status == 'EN ATTENTE COMMERCIALISATION' && lessor_status == 'EN ATTENTE PUBLICATION')
      update_is_mad_updated_at if is_mad_changed?
    end
    
    def main_picture
      pictures.first
    end
    
    def soft_delete
      # "todo : soft delete of housing_offer"
      
      return if Rails.env == 'test'
      Hermes::BoClient::OfferStatusTask.new(housing_offer_id: id).execute
    end
    
    def set_log_action_history(action)
      Hermes::LogJob.perform_later(
        code: 'HERMES_HISTORY_LOG',
        action: "HousingOffer #{action}",
        housing_offer_id: self.id.to_s,
        current_user_id: Hermes::HousingOffer.current_user_id&.to_s
      )
      Hermes::HousingOffer.current_user_id = nil
    end
    
    def assign_target_objective
      # self.target_objective_ids=[]
      self.target_objectives = lessor.get_target_objectives_for(self)
    end
    
    def update_housing_channel_keys!(key, add = true)
      action = :delete
      if add
        raise Hermes::Errors::HousingChannelError, I18n.t('errors.housing_channel.already_added', key: key) if housing_channel_keys.include?(key)
        raise Hermes::Errors::HousingChannelError, I18n.t('errors.housing_channel.unavailable_for_departments', key: key, departments: department, count: 1) if !should_add_channel_key?(key)
        
        action = :push
      end
      housing_channel_keys.send(action, key)
      save!
    end
    
    def should_add_channel_key?(key)
      settings = Hermes::HousingChannelSetting.where(department: department)
      return settings.select {|setting| setting.housing_channel_keys.include? key}.any? if settings.any?
      
      true
    end
    
    def reject_others_application
      Hermes::HousingApplication.where(housing_offer_id: id).each do |appli|
        if appli.status == "EN ATTENTE"
          appli.status = "NON RETENUE"
          appli.is_active = false
          appli.save
        end
      end
    end
    
    def before_save_actions
      picture_urls_format
      manage_housing_channel_keys
      compute_effort_rate if effort_rate.nil?
      udapte_city if district_changed?
      udapte_siret_siren
      if lessor_siren
        self.lessor = Hermes::Lessor.find_by(siren: lessor_siren)
      end
      update_extras_presence
      if entities_siret_changed? || lessor_siren_changed?
        handle_entities_relationships
      end
      inherit_lesssor_data if lessor_id_changed?
      update_department if postal_code_changed? && department.blank?
      update_individual_heating if heating_type_changed?
      update_lessor_rights if lessor_id_changed?
      update_charges if (
      rental_charges_changed? ||
        parking_charges_changed? ||
        typology_changed? ||
        heating_type_changed?)
      if rent_with_charges_changed? || min_effort_rate_changed?
        update_min_revenues
      end
      update_family_size if typology_changed? || accept_low_occupation_changed?
      update_threshold if threshold_type_changed?
      if threshold_id_changed? || threshold_subtype_changed?
        update_own_thresholds
      end
      
      if RENTED_STATUSES.include?(lessor_status)
        lessor_status = "LOUE"
      end
      # TODO: handle is_mad for status
      handle_is_mad if lessor_status_changed? || offer_status_changed?
      update_reserved
      update_eligbility if (changes.keys & ELIGIBILITY_FIELDS).length.positive?
      update_eligible_request_nb if eligibility_query_changed?
      target_objective_computation_needed = (agency_id_changed? ||
        lessor_id_changed? ||
        operation_name_changed? ||
        district_changed? ||
        department_changed? ||
        threshold_type_changed? ||
        new_offer_changed? ||
        priority_public_changed? ||
        typology_changed?)
      assign_target_objective if (lessor && target_objective_computation_needed && (!lessor_status_changed? && !RENTED_STATUSES.include?(lessor_status)))
      
      handle_requalification_status
      if active_reservations_changed?
        update_reservations_readers 
        update_active_reservations_rex_ids
      end
      compute_without_charges if rent_amount.blank?
    end

    def manage_housing_channel_keys
      if housing_channel_keys.include?('dc') && housing_channel_keys.length != 1
        self.housing_channel_keys = self.housing_channel_keys - ['dc']
      elsif housing_channel_keys.length == 0
        self.housing_channel_keys = ['dc']
      end 
    end 

    def update_reserved
      self.reserved = !self.active_reservations&.empty?
    end 

    def get_reader_group_id(entities)
      group_ids = []
      entities.each do |id|
        if entity = Hermes::Entity.find(id)
          group_ids += ['g_'+entity.group_id, 'g_'+entity.supervisor_group_id, 'g_'+entity.manager_group_id ]
        end
      end
      group_ids
    end 

    def update_reservations_readers
      new_entities = (active_reservations || []) - (active_reservations_was || [])
      removed_entities = (active_reservations_was || []) - (active_reservations || [])

      self.readers = (self.readers + get_reader_group_id(new_entities) - get_reader_group_id(removed_entities)).uniq
    end 
    
    def handle_requalification_status
      if ATTRIBUTION_STATUSES.include?(offer_status)
        self.requalification = true
      end
    end
    
    def udapte_city
      self.city = self.district
    end
    
    def update_lessor_rights
      
      if lessor_id_was && lessor_was = Hermes::Lessor.find(lessor_id_was)
        readers.delete("g_#{lessor_was.group_id}")
        writers.delete("g_#{lessor_was.supervisor_group_id}")
        writers.delete("g_#{lessor_was.manager_group_id}")
      end
      
      if lessor
        self.readers = (self.readers || []).push("g_#{lessor.group_id}").uniq
        self.writers = (self.writers || []).push("g_#{lessor.supervisor_group_id}").uniq
        self.writers = (self.writers || []).push("g_#{lessor.manager_group_id}").uniq
      end
    end
    
    def update_active_reservations_rex_ids
      self.active_reservations_rex_ids = [] if !self.active_reservations_rex_ids
      
      new_entities = (active_reservations || []) - (active_reservations_was || [])
      removed_entities = (active_reservations_was || []) - (active_reservations || [])
      
      new_entities.each do |entity_id|
        entity = Hermes::Entity.find(entity_id)
        self.active_reservations_rex_ids.push(entity.id_rex) if entity && !entity.id_rex.blank?
      end
      
      removed_entities.each do |entity_id|
        entity = Hermes::Entity.find(entity_id)
        if entity && !entity.id_rex.blank? && self.active_reservations_rex_ids.include?(entity.id_rex)
          self.active_reservations_rex_ids.delete(entity.id_rex)
        end
      end
    end
    
    def picture_urls_format
      if picture_urls_changed? && !picture_urls.blank?
        
        if picture_urls.is_a? Array
          picture_urls.each do |picture_url|
            if !( (picture_url.is_a? Hash) && (picture_url.key?("full_size")||picture_url.key?("thumb240") ||picture_url.key?("thumb1000")) )
              errors.add :picture_urls, "Ce champ doit etre renseigné avec un tableau d'objet du type : 
                [
                  { full_size: '......',
                    thumb240 : '........',
                    thumb1000 : '........'
                  },
                  {.....},
                  ...
                ]"
            end
          end
        else
          errors.add :picture_urls, "Ce champ doit etre renseigné avec un tableau d'objet du type : 
          [
            { full_size: '......',
              thumb240 : '........',
              thumb1000 : '........'
            },
            {.....},
            ...
          ]"
        end

      end 
    end
    def update_lessor_objectives
      lessor&.handle_objectives_for(self)
    end
    
    def destroy_tomorrow_if_no_update
      AutoDestroyOfferJob.set(wait: 36.hours).perform_later(id.to_s)
    end
    
    def after_create_actions
      set_log_action_history('created') if Hermes::HousingOffer.current_user_id
    end
    
    def after_save_actions
      check_status_changed
      update_lessor_objectives if lessor_status_changed? && RENTED_STATUSES.include?(lessor_status)
      
      update_lessor_objectives if lessor_status_changed? && RENTED_STATUSES.include?(lessor_status)
      destroy_tomorrow_if_no_update if auto_destroy_next_day && last_update_by_lessor_changed?
      
      soft_delete if offer_status_changed? && offer_status == "RETIRE PAR LE BAILLEUR"
      
      if PUBLISHED_STATUSES.include?(offer_status)
        if !PUBLISHED_STATUSES.include?(offer_status_was) || (updated_at.year != updated_at_was&.year)
          Hermes::Accounting.create(year: Time.new.year, target: "offer", object_id: self.id)
        end
      end

      ExecuteFlux5TaskJob.perform_later(self.id.to_s) if is_mad && self.changed?
      
      handle_unpublished_status
      
      set(published: PUBLISHED_STATUSES.include?(offer_status))
      
      update_eligibility_score if eligibility_query_changed?
      run_flux_11_side_effects if need_flux_11_side_effects?
      run_flux_13_side_effects if need_flux_13_side_effects?
      update_readers_writers_application if readers_changed? || writers_changed?
    end

    def send_flux5_sync
      Log.perform_later(level: 'info', domain: 'api', code: "HERMES_FLUX5_SYNC", action: "housing_offer : #{self.id} -- Envoi du flux 5", data: self.attributes.slice(:_id, :als_id, :external_ref, :lessor_status, :offer_status, :lessor_status_id).transform_values{|v| v.to_s})
      res = ExecuteFlux5TaskJob.perform_now(self.id.to_s)
      if res
        Log.perform_later(level: 'info', domain: 'api', code: "HERMES_FLUX5_SYNC", action: "housing_offer : #{self.id} -- Flux 5 envoyé", data: res.attributes.transform_values{|v| v.to_s})
      else
        if xml_sopra_error_ids_changed? && xml_sopra_error_ids.any?
          error = Hermes::XmlSopraError.find(xml_sopra_error_ids.last)&.inspect
        else
          error = "Impossible d'identifier l'erreur, vérifier les codes de log BO_CLIENT_POST et BO_CLIENT_PUT pour l'offre en question"
        end
        Log.perform_later(level: 'error', domain: 'api', code: "HERMES_FLUX5_SYNC_ERROR", action: "housing_offer : #{self.id} -- Erreur d'envoi du flux 5", data: { "error" => error } )
      end
    end

    def update_readers_writers_application
      new_readers = (readers_was || []) - (readers || [])
      old_readers = (readers || []) - (readers_was || [])
      new_writers = (writers_was || []) - (writers || [])
      old_writers = (writers || []) - (writers_was || [])
      
      Hermes::HousingApplication.where(housing_offer_id: id).each do |app|
        app.readers = (app.readers || []) + (new_readers || []) - (old_readers || [])
        app.writers = (app.writers || []) + (new_writers || []) - (old_writers || [])
        app.save
      end
    end
    
    def after_update_actions
      update_affiliated_scores if department_changed?
      set_log_action_history('updated') if Hermes::HousingOffer.current_user_id
    end
    
    def update_affiliated_scores
      ToucheScoresJob.perform_later(
        Hermes::Score.only(:id, :housing_offer_id)
          .where(housing_offer_id: id)
          .pluck(:id)
          .map(&:to_s)
      )
    end
    
    def update_eligibility_score
      UpdateScoreFromOfferJob.perform_later(id.to_s)
    end
    
    def for_single_parent
      if single_min_family_size > max_family_size
        [max_family_size]
      elsif single_min_family_size != max_family_size
        (single_min_family_size..max_family_size).map(&:to_s)
      else
        [single_min_family_size]
      end
    end
    
    def update_lessor_on_all_offers
      offs = Hermes::HousingOffer.all
      counter = 0
      offs.each do |off|
        if (off.lessor_siren)
          off.lessor = Hermes::Lessor.find_by(siren: off.lessor_siren)
          off.save
          counter = counter + 1
        end
      end
    end
    
    def update_eligbility
      # boolean filters
      boolean_filters = %i[
        single_parent
        priority_public
        mutation
      ]
      h = attributes.select do |key, val|
        boolean_filters.include?(key.to_sym) && val
      end.each_with_object({}) do |(key, val), h|
        h[key.to_sym] = val unless val.nil?
      end
      
      # reserved filters reserve field => 'reserved'
      
      if reserved
        h[:$and] ||= []
        h[:$and] << {entity_id: {'$in' => active_reservations}}
        h[:$and] << {entity_validated: true}
      end
      
      # category filter
      h[:$and] ||= []
      h[:$and] << {
        :$or => (1..12).map do |category|
          {
            category_type: category,
            reference_tax_revenues: {
              '$lte': get_tax_threshold(category),
              '$gte': get_tax_threshold(category, true)
            }
          }
        end
      }
      
      # typology filter
      single_min = single_min_family_size
      couple_min = couple_min_family_size
      max_size = max_family_size
      if max_size
        max_size = Integer(max_size =~ /\+$/ ? max_size.sub('+', '') : max_size)
      end
      single_min = single_min > max_size ? max_size : single_min
      
      if single_min && couple_min && max_size
        h[:$and] ||= []
        first = {
          single_parent: true,
          family_size: {:$gte => single_min}
        }
        first[:family_size][:$lte] = max_size if max_size < 12
        first[:family_size][:$ne] = 8 if rooms == 6
        second = {
          single_parent: {:$ne => true},
          family_size: {:$in => for_single_parent.map(&:to_i)}
        }
        h[:$and] << {:$or => [first, second]}
      end
      
      # revenues filter
      h[:monthly_revenues] = {:$gte => min_revenues} if min_revenues.present?
      
      # geo filter
      h[:truly_seeked_departments] = department if department.present?
      
      # housing_channels filter
      h[:housing_channel_keys] = {:$in => housing_channel_keys}
      
      h[:typology_range] = typology
      
      query = h.to_json
      if Hermes::HousingOffer.possible_eligibility_statuses.include?(offer_status)
        self.eligibility_query = query if query != eligibility_query
      else
        self.eligibility_query = nil
      end
      
      h
    end
    
    def get_tax_threshold(cat = nil, for_floor = false)
      prefixe = for_floor ? 'floor_category_' : 'ceil_category_'
      send("#{prefixe}#{cat}")
    end
    
    class << self
      def rented_statuses
        RENTED_STATUSES
      end
      
      def published_statuses
        PUBLISHED_STATUSES
      end
      
      def possible_eligibility_statuses
        POSSIBLE_ELIGIBILITY_STATUSES
      end
      
      def search_by_postal_codes(postal_codes)
        weights = postal_codes.uniq.reverse.each_with_index.map do |postal_code, idx|
          {'$cond' => [{'$eq' => ['$postal_code', postal_code]}, (idx + 1), 0]}
        end
        project = fields.keys.map do |field|
          {field => 1}
        end.reduce({}, :merge)
        
        collection.aggregate([
                               {'$project' => project.merge('weight': {'$add': weights})},
                               {'$sort' => {'weight': -1}}
                             ])
      end
    end
    
    def ordered_pictures
      (ordered_picture_ids || []).map {|picture_id| pictures.find_by(id: picture_id)}
    end
    
    def add_picture(pic_id)
      self.ordered_picture_ids ||= []
      self.ordered_picture_ids << pic_id
      save
    end
    
    def remove_picture(pic_id)
      if ordered_picture_ids
        self.ordered_picture_ids.delete(pic_id)
        save
      end
    end

    def need_flux_11_side_effects?
      lessor_status == 'RENDU AU BAILLEUR CONFIRME' && offer_status == 'RENDU AU BAILLEUR'
    end
    
    def run_flux_11_side_effects
      Flux11SideEffectsJob.perform_later(id.to_s)
    end
    
    def need_flux_13_side_effects?
      lessor_status == 'RETIRE' && (offer_status == 'RETIRE PAR LE BAILLEUR' || offer_status == 'RETIRE PAR ALS')
    end
    
    def run_flux_13_side_effects
      Flux13SideEffectsJob.perform_later(id.to_s)
    end
  end
end
