# frozen_string_literal: true

module Hermes
  class ManagementArea
    include Hermes::Concerns::CommonFieldsConcern
    
    field :name, type: String
    field :code, type: String, default: 'ALSSG'

    include Slugable
    sluggify :code
    
    belongs_to(
        :lessor,
        class_name: 'Hermes::Lessor',
        index: true
    )

    validates_presence_of :name
    owner_field :user_id

    common_indexes *%i[
      name
      code
      slug
    ]

    before_save     :before_save_actions

    def before_save_actions
      self.code = slug != code.downcase ? slug.upcase : code.upcase
      true
    end

  end
end
