# frozen_string_literal: true

module Hermes
  # HousingRequest serializer
  class HousingRequestSerializer < Hermes::ApplicationSerializer
    cache key: proc {"#{object.cache_key}"}, expires_in: 1.day
    type 'housing_requests'

    attribute :backoffice_status
    attribute :is_active
    attribute :in_progress
    attribute :als_id
    attribute :request_status
    attribute :attributed_housing_district
    attribute :attributed_housing_district_postal_code
    attribute :old_request_status
    attribute :request_status_updated_at
    attribute :old_entity_status
    attribute :entity_status_updated_at
    attribute :entity_status
    attribute :dalo
    attribute :asked_mobility
    attribute :valid_mobility
    attribute :decider
    attribute :submitted_to_cal_nb
    attribute :sent_to_cal_nb
    attribute :rejected_in_cal_nb
    attribute :rejected_after_cal_nb
    attribute :accepted_in_cal_nb
    attribute :mobility
    attribute :seeked_types
    attribute :entity_help_required
    attribute :contract_type
    attribute :registration_number
    attribute :applicant_ids
    attribute :entity_validated
    attribute :priority_points_indexes
    attribute :has_priority
    attribute :nud
    attribute :nb_of_requests_made
    attribute :reference_tax_revenues
    attribute :seeked_districts
    attribute :category_type
    attribute :single_parent
    attribute :id_counter
    attribute :id_unique
    attribute :id_unique_associate
    attribute :internal_ref
    attribute :forcing_duplicate
    attribute :electronical_renewal
    attribute :ec_law_ressources
    attribute :is_long_delay_anormal
    attribute :date_calculator_indicator
    attribute :with_car_park
    attribute :widen_request
    attribute :housing_type_list
    attribute :expulsion_judgement_date
    attribute :other_reason
    attribute :work_contract_type
    attribute :seeked_district_postal_codes
    attribute :seeked_departments
    attribute :seeked_localisation_list
    attribute :dependents
    attribute :monthly_revenues
    attribute :dependents_monthly_incomes
    attribute :dependents_monthly_charges
    attribute :typology_range
    attribute :entity_validated
    attribute :family_size
    attribute :coapplicants
    attribute :lessor_name
    attribute :lessor_siren
    attribute :social_housing_entry_date
    attribute :social_housing_name
    attribute :test_created_at
    attribute :sne_number
    attribute :employee_id
    attribute :employee_start_date
    attribute :entity_access_level
    attribute :employee_unit
    attribute :tab_state
    attribute :motivation1
    attribute :motivation2
    attribute :motivation3
    attribute :seeking_first_floor
    attribute :seeking_elevator
    attribute :seeked_max_rent
    attribute :seeked_epci_first_quartils
    attribute :seeked_quartil_belongings
    attribute :sne_creation_date
    attribute :sne_created_at
    attribute :precisions
    attribute :has_multiple_attributions_under_conditions
    attribute :seeked_housing_categories
    attribute :radiation_date
    attribute :radiation_motive
    attribute :min_surface
    attribute :main_applicant_id
    attribute :main_applicant
    attribute :requester_firstname
    attribute :requester_lastname
    attribute :requester_fullname
    attribute :requester_email
    attribute :requester_birth_date
    attribute :coapplicant_ids
    attribute :coapplicants
    attribute :applications_nb
    attribute :last_application_date
    attribute :housing_application_ids
    attribute :active_application_id
    attribute :active_application_housing_offer_id
    attribute :form_status
    attribute :profile_epci_score
    attribute :applicated_position
    attribute :eligible_position
    attribute :effort_rate
    attribute :entity_id
    attribute :documents_required
    attribute :documents_uploaded
    attribute :unborn_children
    attribute :is_house_owner
    attribute :workplace_insee_code
    attribute :unique_numbers
    attribute :providers
    attribute :unique_numbers_complete
    attribute :is_untensed_area
    attribute :housing_channel_chosen_keys
    attribute :housing_channel_pending_keys
    attribute :housing_channel_keys
    attribute :is_sne_update_available
    attribute :last_search_on_sne
    attribute :housing_offer_ids
    attribute :last_search_on_sne
    attribute :viewed_by_colaborator_offre_logement
    if HERMES_CONFIG['DEBUG_ERRORS'] == false
      attribute :last_xml_validation_report
    end
    attribute :last_send_has_error
    attribute :last_entity_code
    attribute :consent_level
    attribute :cdii
    attribute :last_update_user_id
    attribute :last_update_user_last_name
    attribute :last_update_user_first_name


    attribute :dependents, key_transformation: :unaltered
    attribute :disability_details, key_transformation: :unaltered
    attribute :profession_details, key_transformation: :unaltered
    attribute :actual_housing_details, key_transformation: :unaltered
    attribute :postal_details, key_transformation: :unaltered
    attribute :owned_housing_details, key_transformation: :unaltered
    attribute :applicant_verified
    
    attribute :eligible_offers_nb_detailed, if: :detailed_eligibility_count?
    
    belongs_to :housing_request_folder, serializer: HousingRequestFolderSerializer do
      object.housing_request_folder
    end
    
    def housing_offer_ids
      object.housing_offer_ids&.map(&:to_s) || []
    end
    
    def unique_numbers
      object.unique_numbers
    end
    
    def providers
      res = []
      object.nuds&.each do |nud|
        res.push(nud.provider)
      end
      res
    end
    
    def unique_numbers_complete
      res = []
      object.nuds&.each do |nud|
        res.push({id: nud.id.to_s, unique_number: nud.unique_number, provider: nud.provider} )
      end
      res
    end
    
    def form_status
      object.form_status
    end
    
    def detailed_eligibility_count?
      instance_options[:detailed_eligibility] == true
    end
    
    def eligible_offers_nb_detailed
      return "uneligible" unless object.main_applicant&.eligibility_query.present?
      
      eligibility_json = object.main_applicant.eligibility_query.json_to_h
      eligible_offers = Hermes::HousingOffer.where(eligibility_json)
      {
        non_commercialized: eligible_offers.only(:commercialized)
                              .where(commercialized: {:$ne => true})
                              .count,
        commercialized_publicated: eligible_offers.only(
          :commercialized,
          :is_mad
        ).where(
          commercialized: true,
          is_mad: true
        ).count,
        commercialized_non_publicated: eligible_offers.only(
          :commercialized,
          :is_mad
        ).where(
          commercialized: true,
          is_mad: {:$ne => true}
        ).count
      }
    end
    
    def score
      return unless instance_options[:scores]
      
      @score ||= instance_options[:scores].select do |score|
        score.applicant_id == object.main_applicant_id
      end.first
    end
    
    def handle_score_value(value)
      value.nil? ? 'uneligible' : value
    end
    
    def profile_epci_score
      handle_score_value score&.profile_epci_score
    end
    
    def applicated_position
      handle_score_value score&.applicated_position
    end
    
    def eligible_position
      handle_score_value score&.eligible_position
    end
    
    def effort_rate
      handle_score_value score&.effort_rate
    end
    
    def main_applicant_id
      object.main_applicant_id&.to_s
    end
    
    def main_applicant
      return nil unless object.main_applicant
      
      Hermes::ApplicantSerializer.new(
        object.main_applicant,
        scope: scope,
        root: false
      )
    end
    
    def entity_id
      object.entity_id&.to_s
    end
    
    def coapplicant_ids
      object.applicant_ids&.map(&:to_s)
    end
    
    def coapplicants
      object.coapplicants
    end
    
    def housing_application_ids
      object.housing_applications&.map {|a| a.id.to_s} || []
    end
    
    def housing_applications
      object.housing_applications.map do |application|
        Hermes::HousingApplicationSerializer.new(
          application,
          scope: scope,
          root: false
        )
      end
    end
  end
end
