require 'rails_helper'

RSpec.describe(Hermes::EntityValidation, type: :model) do
  it 'has a valid factory' do
    expect(FactoryBot.create(:entity_validation)).to be_valid
  end

  describe 'Hooks' do
    let!(:entity) { create(:entity) }
    let!(:entity_validation) { create(:entity_validation, entity: entity) }

    before(:each) do
      drain_all_jobs
    end

    it 'should set auto validation' do
      expect(entity_validation.validation).to be(entity.auto_validation)
    end

    it 'should active' do
      expect(entity_validation.active).to be(true)
    end

    it 'should inactive if dates are outdated' do
      entity_validation.update!(contract_end_date: Date.today - 1.days)
      expect(entity_validation.reload.active).to be(false)
    end

    it 'should call close contract job' do
      close_date = Date.today
      entity_validation.update!(contract_end_date: close_date)
      job = enqueued_jobs.find{ |j| j[:job] == CloseContractJob}
      expect(job[:args]).to include(entity_validation.id.to_s)
    end
  end
end
